@@ -25,6 +25,8 @@
 #  include <config.h>
 #endif
 
+#include <stdlib.h>
+#include <limits.h>
 #include <sys/types.h>
 #include <dirent.h>
 #include <signal.h>
@@ -42,6 +44,9 @@
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
 #include <X11/Xatom.h>
+#ifdef HAVE_XEVIE
+#include <X11/extensions/Xevie.h>
+#endif
 #include "main.h"
 #include "gok.h"
 #include "gok-word-complete.h"
@@ -64,6 +69,14 @@
 #include "gok-keyslotter.h"
 #include "gok-gconf.h"
 
+#ifdef ENABLE_LIBUSB_INPUT
+#include "gok-libusb.h"
+#endif
+
+extern int errno;
+
+static Display *xevie_dpy = NULL;
+
 #define USE_GCONF 1
 
 #define APP_STATIC_BUFF_SZ 30
@@ -77,7 +90,8 @@
 #define GOK_ERROR_ACTION_OPEN         6
 #define GOK_ERROR_FEEDBACK_OPEN       7
 #define GOK_ERROR_NO_XKB_EXTENSION    8
-#define GOK_ERROR_NOSTICKYKEYS		  9
+#define GOK_ERROR_NOSTICKYKEYS	      9
+#define GOK_ERROR_LIBUSB              10
 
 #define GOKMAINKEYBOARDNAME "main"
 #define GOKLOGINKEYBOARDNAME "Keyboard"
@@ -88,8 +102,6 @@
 
 static gboolean respawn_on_segv = TRUE;
 
-static gboolean queue_input_events = TRUE;
-
 static guint input_idle_handler = 0;
 
 static GokApplication *_gok_app = NULL;
@@ -123,7 +135,7 @@ static gint m_GeometryY = -1;
 /* pointer to the gok command predictor */
 static Gok_Predictor m_pCommandPredictor = NULL;
 
-/* poptOption storage structure */
+/* command line aguments storage structure */
 typedef struct _GokArgs {
 	char* accessmethodname;
 	char* inputdevicename;
@@ -141,6 +153,7 @@ typedef struct _GokArgs {
 	char* selectactionname;
 	gboolean list_actions;
 	gboolean list_accessmethods;
+	char* usb_vid_pid;
 	gdouble valuator_sensitivity;
 } GokArgs;
 
@@ -162,6 +175,7 @@ static void gok_main_object_state_listener (Accessible* pAccessible);
 static gboolean gok_main_has_xkb_extension ();
 static gboolean gok_main_check_sticky_keys (GtkWidget *widget);
 void gok_main_mouse_button_listener (gint button, gint state, long mods, long timestamp);
+static gboolean gok_main_xevie_io (GIOChannel *source, GIOCondition condition, gpointer data);
 
 /* extern funcs */
 void       gok_input_detach_corepointer (void);
@@ -169,84 +183,87 @@ void       gok_input_restore_corepointer (void);
 
 static gint _screen_height, _screen_width;
 
+#ifdef ENABLE_LIBUSB_INPUT
+static gok_libusb_t *libusb_handle;
+#endif
 
-static struct poptOption options[] = {
+static const GOptionEntry options[] = {
 	{
 		"access-method", 
 		'a',
-		POPT_ARG_STRING, 
-		&gok_args.accessmethodname, 
 		0,
+		G_OPTION_ARG_STRING, 
+		&gok_args.accessmethodname, 
 		N_("Use the specified access method. NAME is a string and can be found in the various access method files (.xam) assigned to the \"name\" property of <gok:accessmethod> tag. Note this is not necessarily the same as the name of the .xam file. (See --list-accessmethods)"), 
 		N_("NAME")
 	},
 	{
 		"editor", 
-		'e', 
-		POPT_ARG_NONE, 
+		'e',
+		0,	
+		G_OPTION_ARG_NONE, 
 		&gok_args.display_keyboard_editor, 
-		0,
 		N_("Start the GOK keyboard editor"), 
 		NULL},
 	{
 		"extras", 
 		'\0', 
-		POPT_ARG_NONE, 
-		&gok_args.use_extras, 
 		0,
+		G_OPTION_ARG_NONE, 
+		&gok_args.use_extras, 
 		N_("Use special, but possibly unstable, gok stuff"),
 		NULL},
 	{
 		"geometry", 
 		'\0', 
-		POPT_ARG_STRING, 
-		&gok_args.geometry, 
 		0,
+		G_OPTION_ARG_STRING, 
+		&gok_args.geometry, 
 		N_("Whenever --geometry is not used gok remembers its position between invocations and starts in the position that it had when it was last shutdown.  When --geometry is used gok positions itself within the rectangular area of screen described by the given X11 geometry specification.  When --geometry is used gok does not remember its position when it shuts down.  This behaviour can be changed with the --remembergeometry flag which forces gok to remember its position when shutdown even when it was started with --geometry."),
 		N_("GEOMETRY")
 	},
 	{
 		"input-device", 
 		'i', 
-		POPT_ARG_STRING, 
-		&gok_args.inputdevicename, 
 		0,
+		G_OPTION_ARG_STRING, 
+		&gok_args.inputdevicename, 
 		N_("Use the specified input device"), 
 		N_("DEVICENAME")
 	},
 	{
 		"keyboard", 
 		'k', 
-		POPT_ARG_STRING, 
-		&gok_args.mainkeyboardname, 
 		0,
+		G_OPTION_ARG_STRING, 
+		&gok_args.mainkeyboardname, 
 		N_("Start GOK with the specified keyboard."), 
 		N_("KEYBOARDNAME")
 	},
 	{
 		"list-accessmethods", 
 		'\0', 
-		POPT_ARG_NONE, 
-		&gok_args.list_accessmethods, 
 		0,
+		G_OPTION_ARG_NONE, 
+		&gok_args.list_accessmethods, 
 		N_("List the access methods that can be used as options to other arguments."),
 		NULL
 	},
 	{
 		"list-actions", 
 		'\0', 
-		POPT_ARG_NONE, 
-		&gok_args.list_actions, 
 		0,
+		G_OPTION_ARG_NONE, 
+		&gok_args.list_actions, 
 		N_("List the actions that can be used as options to other arguments."),
 		NULL
 	},
 	{
 		"login", 
 		'l', 
-		POPT_ARG_NONE, 
-		&gok_args.is_login, 
 		0,
+		G_OPTION_ARG_NONE, 
+		&gok_args.is_login, 
 		N_("GOK will be used to login"),
 		NULL
 	},
@@ -254,9 +271,9 @@ static struct poptOption options[] = {
 	{
 		"debugnameless", 
 		'\0', 
-		POPT_ARG_NONE, 
-		&gok_args.debug_nameless, 
 		0,
+		G_OPTION_ARG_NONE, 
+		&gok_args.debug_nameless, 
 		N_("GOK will display keys for nameless GUI. Useful for debugging."),
 		NULL
 	},
@@ -264,9 +281,9 @@ static struct poptOption options[] = {
 	{
 		"remembergeometry", 
 		'\0', 
-		POPT_ARG_NONE,
-		&gok_args.remember_geometry, 
 		0,
+		G_OPTION_ARG_NONE,
+		&gok_args.remember_geometry, 
 		N_("Can be used with --geometry.  Forces GOK to remember its position when shutdown even when it was started with --geometry.  Please see the discussion under the --geometry flag for more information."),
 		NULL
 	},
@@ -275,9 +292,9 @@ static struct poptOption options[] = {
 		/* note: not enough info for 5-switch directed scanning */
 		"scan-action", 
 		'\0', 
-		POPT_ARG_STRING, 
-		&gok_args.scanactionname, 
 		0,
+		G_OPTION_ARG_STRING, 
+		&gok_args.scanactionname, 
 		N_("Start GOK and hook this action to scan operations. (See --list-actions)"), 
 		NULL
 	},
@@ -285,33 +302,43 @@ static struct poptOption options[] = {
 		/* primarily for use at login since there is no gconf */
 		"select-action", 
 		'\0', 
-		POPT_ARG_STRING, 
-		&gok_args.selectactionname, 
 		0,
+		G_OPTION_ARG_STRING, 
+		&gok_args.selectactionname, 
 		N_("Start GOK and hook this action to select operations. (See --list-actions)"), 
 		NULL
 	},
 	{
 		"settings", 
 		's', 
-		POPT_ARG_NONE, 
-		&gok_args.display_settings_dialog, 
 		0,
+		G_OPTION_ARG_NONE, 
+		&gok_args.display_settings_dialog, 
 		N_("Open the settings dialog box when GOK starts"),
 		NULL
 	},
+#ifdef ENABLE_LIBUSB_INPUT
+	{
+		"use-libusb-for-input", 
+		'\0',
+		0, 
+		G_OPTION_ARG_STRING,
+		&gok_args.usb_vid_pid,
+		N_("Use libusb for input events and use the device with the specified USB Vendor ID (VID) and Product ID (PID)."),
+		"VID:PID"
+	},
+#endif
 	{
 		"valuator-sensitivity", 
 		'\0', 
-		POPT_ARG_DOUBLE, 
-		&gok_args.valuator_sensitivity, 
 		0,
+		G_OPTION_ARG_DOUBLE, 
+		&gok_args.valuator_sensitivity, 
 		N_("A multiplier to be applied to input device valuator events before processing"),
 		NULL
 	},
-
 	/* End the list */
-	{NULL, '\0', 0, NULL, 0, NULL, NULL}
+	{	NULL}
 };
 
 static int segfaults = 0;
@@ -326,8 +353,9 @@ static int gok_sig_handler(int sig)
 #if ! defined ENABLE_LOGGING_NORMAL
 		if (respawn_on_segv)
 		{
-		    fprintf (stderr, "Restarting GOK.\n"); 
-		    g_spawn_command_line_async ("gok", NULL); 
+		    fprintf (stderr, "Restarting GOK.\n");
+		    g_on_error_stack_trace("gok");
+		    g_spawn_command_line_async ("gok", NULL);
 		}
 #endif
 		break;
@@ -348,7 +376,7 @@ static int gok_sig_handler(int sig)
 static void
 gok_args_init (GokArgs *args)
 {
-	/* initialize command line option (popt) storage */
+	/* initialize command line option storage */
 	args->accessmethodname = NULL;
 	args->inputdevicename = NULL;
 	args->mainkeyboardname = NULL;
@@ -363,6 +391,7 @@ gok_args_init (GokArgs *args)
 	args->use_extras = FALSE;
 	args->scanactionname = NULL;
 	args->selectactionname = NULL;
+	args->usb_vid_pid = NULL;
 	args->valuator_sensitivity = 0;
 }
 
@@ -468,7 +497,7 @@ gint main (gint argc, gchar *argv[])
 
 	/* intialization */
 	result = gok_main_open(argc, argv);
-	
+		
 	if (result != 0)
 	{
 		return result;
@@ -536,6 +565,9 @@ gint
 gok_main_open(gint argc, gchar *argv[])
 {
 	GnomeProgram *gok_program;
+	GOptionContext *context;
+	gboolean create_dirs = TRUE;
+	int i;
 
 	/* initialize member data */
 	m_pForegroundWindowAccessible = NULL;
@@ -543,19 +575,35 @@ gok_main_open(gint argc, gchar *argv[])
 	m_pWindowMain = NULL;
 	m_pCurrentKeyboard = NULL;
 	m_countIgnoreConfigure = 0;
+	struct gok_window_realize_data *window_realize_data = g_new0(struct gok_window_realize_data, 1);
 
 	gok_args_init (&gok_args);
 
+	context	= g_option_context_new (_("The GNOME On-screen Keyboard"));
+	g_option_context_add_main_entries (context, options, GETTEXT_PACKAGE);
+
+	/*
+	 * Preprocess the arguments to see if --login was included.  Need to
+	 * do this before calling gnome_program_init (which normally processes
+	 * the arguments) since this affects how we call gnome_program_init.
+	 */
+	for (i = 0; i < argc; i++) {
+		if (strcmp (argv[i], "--login") == 0) {
+			create_dirs = FALSE;
+			break;
+		}
+	}
+
 	/* Initialize Gnome */
 	gok_program = gnome_program_init (PACKAGE, VERSION,
 		LIBGNOMEUI_MODULE, argc, argv,
-		GNOME_PARAM_POPT_TABLE, options,
+		GNOME_PARAM_GOPTION_CONTEXT, context,
 		GNOME_PROGRAM_STANDARD_PROPERTIES,
 	        GNOME_PARAM_APP_DATADIR, DATADIR"/gok", 
 		LIBGNOMEUI_PARAM_CRASH_DIALOG, FALSE,
+		GNOME_PARAM_CREATE_DIRECTORIES, create_dirs,
 		NULL);
 		
-	
 	if (!bonobo_init (&argc, argv))
 	{
 	    if (gok_args.is_login)
@@ -768,13 +816,56 @@ gok_main_open(gint argc, gchar *argv[])
 	gok_spy_register_windowchangelistener ((void *)gok_main_window_change_listener);
 	gok_spy_register_mousebuttonlistener ((void *)gok_main_mouse_button_listener);	
 	gok_spy_register_objectstatelistener ((void *)gok_main_object_state_listener);
-	
+#ifdef HAVE_XEVIE
+	xevie_dpy = XOpenDisplay(NULL);
+	if (!XevieStart (xevie_dpy)) 
+	  xevie_dpy = NULL;
+	else {
+	  GIOChannel *ioc;
+	  int fd;
+	  XevieSelectInput(xevie_dpy, PointerMotionMask);
+	  gok_log ("xevie opened, PointerMotionMask selected: %d\n", (int)xevie_dpy);
+	  fd = ConnectionNumber(xevie_dpy);
+	  ioc = g_io_channel_unix_new (fd);
+	  g_io_add_watch (ioc, G_IO_IN | G_IO_HUP, gok_main_xevie_io, xevie_dpy);
+	  g_io_channel_unref (ioc);
+	}
+#endif	
 	/* initialize sound */
 	gok_sound_initialize();
 
+	/* initialize libusb if it's been selected */
+	window_realize_data->use_libusb = FALSE;
+#ifdef ENABLE_LIBUSB_INPUT
+	if (gok_args.usb_vid_pid != NULL) {
+		 long vid, pid;
+		 gchar **vid_pid = g_strsplit(gok_args.usb_vid_pid, ":", -1);
+	
+		 if (g_strv_length(vid_pid) != 2) {
+		 	g_strfreev(vid_pid);
+		 	gok_main_display_error(_("can't initialize the libusb backend - can't parse USB VID:PID pair"));
+		 	return GOK_ERROR_LIBUSB;
+		 }
+
+		vid = strtoul(vid_pid[0], NULL, 16);
+		pid = strtoul(vid_pid[1], NULL, 16);
+		g_strfreev(vid_pid);
+		
+		if ((vid == 0 || pid == 0) && errno == EINVAL) {
+			gok_main_display_error(_("can't initialize the libusb backend - can't parse USB VID:PID pair"));
+		 	return GOK_ERROR_LIBUSB;
+		}
+		
+		if ((libusb_handle = gok_libusb_init(vid, pid)) == NULL) {
+		 	return GOK_ERROR_LIBUSB;
+		}
+		window_realize_data->use_libusb = TRUE;
+	}
+#endif
+
 	/* create the main window */
-	m_pWindowMain = gok_main_create_window (
-		gok_data_get_dock_type () != GOK_DOCK_NONE);
+	window_realize_data->is_dock = (gok_data_get_dock_type () != GOK_DOCK_NONE);
+	m_pWindowMain = gok_main_create_window (window_realize_data);
 	if (m_pWindowMain == NULL)
 	{
 		gok_main_display_error (_("Can't create the main GOK window!"));
@@ -787,7 +878,7 @@ gok_main_open(gint argc, gchar *argv[])
 	gok_keyboard_initialize();
 	gok_branchbackstack_initialize();
 
-	fprintf (stderr, "login mode = %s\n", gok_args.is_login ? "true" : "false");
+	gok_log ("login mode = %s\n", gok_args.is_login ? "true" : "false");
 	if (!gok_args.is_login) {
 		gok_main_initialize_wordcomplete ();
 		gok_main_initialize_commandprediction ();
@@ -814,7 +905,10 @@ gok_main_open(gint argc, gchar *argv[])
 
 	if (!gok_args.is_login) {
 		gok_main_check_accessibility();
-		gok_main_warn_corepointer (FALSE, FALSE, TRUE);
+		if (gok_args.usb_vid_pid == NULL && xevie_dpy == NULL) {
+		        /* we're not using libusb or xevie */
+			gok_main_warn_corepointer (FALSE, FALSE, TRUE);
+		}
 	}
 	
 	if (!gok_main_check_sticky_keys (m_pWindowMain)) {
@@ -842,6 +936,8 @@ gok_log ("finished check");
 	signal(SIGTERM, (void (*)(int))gok_sig_handler);
 	signal(SIGINT,  (void (*)(int))gok_sig_handler);
 
+	g_object_unref (gok_program);
+
 	return 0;
 }
 
@@ -1123,12 +1219,46 @@ gok_main_object_state_listener (Accessible* pAccessible)
 static gboolean gok_main_idle_input (gpointer data)
 {
     input_idle_handler = 0;
+
     if ((gok_motion_event.n_axes > 0) && (gok_motion_event.n_axes < 4)) {
 	gok_scanner_input_motion (gok_motion_event.axes, gok_motion_event.n_axes);
     }
     return FALSE;
 }
 
+static gboolean gok_main_xevie_io (GIOChannel *source,
+				   GIOCondition condition,
+				   gpointer data)
+{
+    XEvent xevie_event;
+#ifdef HAVE_XEVIE
+    while (xevie_dpy && XPending (xevie_dpy)) {
+      XMotionEvent *motion_event;
+      gboolean has_active_key = gok_scanner_key_is_active ();
+      /* Xevie has a motion event for us to check */
+      XNextEvent (xevie_dpy, &xevie_event);
+      switch (xevie_event.type)
+	{
+	case MotionNotify:
+	  motion_event = (XMotionEvent*) &xevie_event;
+	  gok_scanner_mouse_movement (motion_event->x_root, motion_event->y_root);
+	  break;
+	default:
+	  fprintf (stderr, "that's odd, unsolicited xevie event\n");
+	}
+      if (!has_active_key)
+	XevieSendEvent (xevie_dpy, &xevie_event, XEVIE_UNMODIFIED);
+    }
+#endif
+    return TRUE;
+}
+
+gboolean
+gok_main_attach_new_devices (void)
+{
+  return (xevie_dpy == NULL);
+}
+
 void
 gok_main_queue_input_event (int *motion_data, gint n_axes)
 {
@@ -1156,8 +1286,7 @@ gok_main_queue_input_event (int *motion_data, gint n_axes)
 *
 * This handler is called each time there is a motion event from the connected input device.
 * We queue the input event for processing during idle, as the rate of input events
-* may exceed our ability to process them, otherwise, and we only need to consider the most
-* recent motion event.
+* may exceed our ability to process them.
 */
 void gok_main_motion_listener (gint n_axes, int *motion_data, long mods, long timestamp)
 {
@@ -1171,19 +1300,15 @@ void gok_main_motion_listener (gint n_axes, int *motion_data, long mods, long ti
 			mods, timestamp);
 		gok_log_leave ();
 #endif
-		if (queue_input_events)
-		    gok_main_queue_input_event (motion_data, n_axes);
-		else
-		    gok_scanner_input_motion (motion_data, n_axes); 
+
+		gok_main_queue_input_event (motion_data, n_axes);
 
 		if (gok_data_get_drive_corepointer ()) {
 			Display *display;
 			GdkWindow *root;
 			display = GDK_WINDOW_XDISPLAY (window->window);
 			root = gdk_screen_get_root_window (gdk_drawable_get_screen (window->window));
-			XWarpPointer (display, None, 
-				      GDK_WINDOW_XWINDOW (root),
-				      0, 0, 0, 0, motion_data[0], motion_data[1]);
+			XWarpPointer (display, None, GDK_WINDOW_XWINDOW (root), 0, 0, 0, 0, motion_data[0], motion_data[1]);
 		}
 	}
 }
@@ -1197,7 +1322,7 @@ void gok_main_motion_listener (gint n_axes, int *motion_data, long mods, long ti
 *
 * This handler is called each time there is a button event.
 */
-void gok_main_button_listener (gint button, gint state, long mods, long timestamp)
+void gok_main_button_listener (gint button, gint state, long mods, long timestamp, gboolean using_libusb)
 {
 	gok_log_enter();
 	gok_log("button[%d] state[%d] mods[%lx] timestamp[%ld]",button,state,mods,timestamp);
@@ -1217,8 +1342,7 @@ void gok_main_button_listener (gint button, gint state, long mods, long timestam
 			       &win_x, &win_y, &mask_ret);
 		SPI_generateMouseEvent (x, y, ename);
 	}
-
-	if (gok_input_corepointer_detached ())
+	if (gok_input_corepointer_detached() || using_libusb == TRUE)
 	{
 		gok_main_mouse_button_listener (button, state, mods, timestamp);
 	}
@@ -1702,7 +1826,7 @@ gboolean gok_main_ds (GokKeyboard* pKeyboard)
 *
 * returns: A pointer to the window, NULL if it couldn't be created.
 */
-GtkWidget* gok_main_create_window (gboolean is_dock)
+GtkWidget* gok_main_create_window (struct gok_window_realize_data *window_realize_data)
 {
 	GtkWidget *window1;
 	GtkWidget *fixed1;
@@ -1724,7 +1848,7 @@ GtkWidget* gok_main_create_window (gboolean is_dock)
 				"border_width", 0,
 				"accept_focus", FALSE,
 				NULL),
-			     "signal::realize", on_window1_realize, is_dock,
+			     "signal::realize", on_window1_realize, window_realize_data,
 			     "signal::destroy", on_window1_destroy, NULL,
 			     NULL);
 
@@ -2554,9 +2678,16 @@ gok_main_close()
 	GokKeyboard* pKeyboard;
 	GokKeyboard* pKeyboardTemp;
 
-	/* get rid of input idle handler */
-	if (input_idle_handler) g_source_remove (input_idle_handler);
-
+	
+	/* cleanup the libusb stuff if we were using it */
+	if (gok_args.usb_vid_pid != NULL) {
+#ifdef ENABLE_LIBUSB_INPUT
+		gok_libusb_cleanup(libusb_handle);
+#endif
+	} else {
+		/* get rid of input idle handler */
+		if (input_idle_handler) g_source_remove (input_idle_handler);
+	}
 	/* restore the corepointer, if it was changed... */
 	gok_input_restore_corepointer ();
 
@@ -2703,12 +2834,21 @@ void gok_main_get_geometry (GdkRectangle* pRectangle)
 gboolean
 gok_main_window_contains_pointer (void)
 {
-	gint x, y;
+	gint x=-1, y=-1, w=0, h=0;
+	gint px, py;
 	GdkModifierType mask;
 	GtkWidget *widget = gok_main_get_main_window ();
-	if (gdk_window_get_pointer (widget->window, &x, &y, &mask) != NULL)
-		return TRUE;
-	return FALSE;
+	gok_scanner_get_pointer_location (&px, &py);
+	if (widget->window) {
+	  gdk_window_get_position (widget->window, &x, &y);
+	  x = px-x;
+	  y = py-y;
+	  gdk_window_get_size (widget->window, &w, &h);
+	}
+	if (x < 0 || y < 0 || x > w || y > h)
+	  return FALSE;
+	else
+	  return TRUE;
 }
 
 Display *
